import numpy as np
from infostop import utils

def best_partition(coords, r1=10, r2=10, return_medoid_labels=False):
    """Infer best stop-location labels from stationary points using infomap.

    The method entils the following steps:
        1.  Detect which points are stationary and store only the median (lat, lon) of
            each stationarity event. A point belongs to a stationarity event if it is 
            less than `r1` meters away from the median of the time-previous collection
            of stationary points.
        2.  Compute the pairwise distances between all stationarity event medians.
        3.  Construct a network that links nodes (event medians) that are within `r2` m.
        4.  Cluster this network using two-level Infomap.
        5.  Put the labels back info a vector that matches the input data in size.
    
    Input
    -----
        coords : array-like (N, 2)
        r1 : number
            Max distance between time-consecutive points to label them as stationary
        r2 : number
            Max distance between stationary points to form an edge.
        return_medoid_labels : bool
            If True, return labels of median values of stationary events, not `coords`.

    Output
    ------
        out : array-like (N, ) 
    """
    # PREPROCESS
    # ----------
    # Time-group points
    groups = utils.group_time_distance(coords, r_C=r1)
    
    # Reduce time-grouped points to their median. Only keep stat. groups (size > 1)
    stat_medoids, medoid_map = utils.get_stationary_medoids(groups, min_size=1)
    
    # Compute their pairwise distances
    pairwise_dist = utils.haversine_pdist(stat_medoids)
    
    # NETWORK
    # -------
    # Construct a network where nodes are stationary location events
    # and edges are formed between nodes if they are within distance `r2`
    c = stat_medoids.shape[0]
    mask = pairwise_dist < r2
    
    # Take edges between points where pairwise distance is < r2
    edges = [
        (i, j)
        for i in range(c) for j in range(i+1, c)
        if mask[int(j + i * c - (i+1) * (i+2) / 2)]  # index in `mask` corresponding to i,j in `pairwise_dist`:
    ]                                                # square matrix flattened index of i, j *minus* lower triangle
                                                     # (including diagonal) down to i+1.

    if len(edges) < 1:
        raise Exception("Found only %d edge(s). Provide longer trajectory or increase `r2`.")
        
    # INFER LABELS
    # ------------
    # Infer the partition with infomap. Partiton has from {node: community, ...}
    partition = utils.infomap_communities(edges)
    
    # Cast the partition as a vector of labels like [community, community, ...]
    labels = np.array([
        partition[n] if n in partition else -1
        for n in range(c)]
    )
    
    # Optionally, just return labels of medians of stationary points
    if return_medoid_labels:
        return labels
    
    # POSTPROCESS
    # -----------
    # Label all the input points and return that label vector
    coord_labels = np.ones(shape=coords.shape[0], dtype=int) * (-1)
    for gid, lab in enumerate(labels):
        coord_labels[medoid_map == gid] = lab
    
    return coord_labels


import infomap
import numpy as np

def haversine(points_a, points_b, radians=False):
    """ 
    Calculate the great-circle distance bewteen points_a and points_b
    points_a and points_b can be a single points or lists of points.

    Author: Piotr Sapiezynski
    Source: https://github.com/sapiezynski/haversinevec

    Using this because it is vectorized (stupid fast).
    """
    def _split_columns(array):
        if array.ndim == 1:
            return array[0], array[1] # just a single row
        else:
            return array[:,0], array[:,1]

    if radians:
        lat1, lon1 = _split_columns(points_a)
        lat2, lon2 = _split_columns(points_b)

    else:
    # convert all latitudes/longitudes from decimal degrees to radians
        lat1, lon1 = _split_columns(np.radians(points_a))
        lat2, lon2 = _split_columns(np.radians(points_b))

    # calculate haversine
    lat = lat2 - lat1
    lon = lon2 - lon1
    d = np.sin(lat * 0.5) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(lon * 0.5) ** 2
    h = 2 * 6371e3 * np.arcsin(np.sqrt(d))
    return h  # in meters

def haversine_pdist(points, radians=False):
    """ 
    Calculate the great-circle distance bewteen each pair in a set of points.
    
    Author: Piotr Sapiezynski
    Source: https://github.com/sapiezynski/haversinevec

    Input
    -----
        points : array-like (shape=(N, 2))
            (lat, lon) in degree or radians (default is degree)

    Output
    ------
        result : array-like (shape=(N*(N-1)//2, ))
    """ 
    c = points.shape[0]
    result = np.zeros((c*(c-1)//2,), dtype=np.float64)
    vec_idx = 0
    if not radians:
        points = np.radians(points)
    for idx in range(0, c-1):
        ref = points[idx]
        temp = haversine(points[idx+1:c,:], ref, radians=True)
        result[vec_idx:vec_idx+temp.shape[0]] = temp
        vec_idx += temp.shape[0]
    return result

def group_time_distance(coords, r_C):
    """Group temporally adjacent points if they are closer than r_C.
    
    Input
    -----
        coords : array-like (shape=(N, 2))
        r_C : number (critical radius)
    
    Output
    ------
        groups : list-of-list
            Each list is a group of points
    """
    groups = []
    
    current_group = coords[0].reshape(-1, 2)
    for coord in coords[1:]:
        
        # Compute distance to current group
        dist = haversine(np.median(current_group, axis=0), coord)
    
        # Put in current group
        if dist <= r_C:
            current_group = np.vstack([current_group, coord])
        # Or start new group if dist is too large
        else:
            groups.append(current_group)
            current_group = coord.reshape(-1, 2)

    # Add the last group
    groups.append(current_group)

    return groups

def get_stationary_medoids(groups, min_size=1):
    """Convert groups of multiple points (stationary location events) to median-of-group points.
    
    Input
    -----
        groups : list-of-list
            Each list is a group of points
        min_size : int
            Minimum size of group to consider it stationary (default: 1)
            
    Output
    ------
        stat_coords : array-like (M, 2)
            Medioids of stationary groups
        mediod_map : list
            Group labels
            
    """
    stat_coords = np.empty(shape=(0, 2))
    medoid_map = []
    i = 0
    for g in groups:
        if g.shape[0] > min_size:
            stat_coords = np.vstack([stat_coords, np.median(g, axis=0).reshape(1, -1)])
            medoid_map.extend([i] * len(g)); i += 1
        else:
            medoid_map.append(-1)
     
    return stat_coords, np.array(medoid_map)


def infomap_communities(edges):
    """Two-level partition of single-layer network with Infomap.
    
    Input
    -----
        edges : list of tuples
            Example: `("cat", "dog", 1)` or (0, 1)

    Output
    ------
        out : dict (node-community hash map)
    """
    # Represent node names as indices
    name_map = {}
    name_map_inverted = {}
    for id_, n in enumerate(set([n for e in edges for n in e[:2]])):  # Loop over nodes
        name_map_inverted[id_] = n
        name_map[n] = id_
       
    # Initiate two-level Infomap
    infomapSimple = infomap.Infomap("--two-level")
    network = infomapSimple.network()
    
    # Add links (weighted)
    if len(edges[0]) == 2:
        for n1, n2 in edges:
            network.addLink(name_map[n1], name_map[n2], 1)

    # Add links (unweighted)
    if len(edges[0]) == 3:
        for n1, n2, w in edges:
            network.addLink(name_map[n1], name_map[n2], w)
    
    # Run infomap
    infomapSimple.run()
    
    # Return in node-community dictionary format
    return dict([
        (name_map_inverted[k], v)
        for k, v in infomapSimple.getModules().items()
    ])
